/*
 * Handle unaligned accesses by emulation.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1996 by Ralf Baechle
 *
 * This file contains exception handler for address error exception with the
 * special capability to execute faulting instructions in software.  The
 * handler does not try to handle the case when the program counter points
 * to an address not aligned to a word boundary.
 *
 * Putting data to unaligned addresses is a bad practice even on Intel where
 * only the performance is affected.  Much worse is that such code is non-
 * portable.  Due to several programs that die on MIPS due to alignment
 * problems I decieded to implement this handler anyway though I originally
 * didn't intend to do this at all.
 *
 * For now I enable fixing of address errors by default to make life easier.
 * I however intend to disable this somewhen in the future when the alignment
 * problems with user programs have been fixed.  For programmers this is the
 * right way to go.
 *
 * Fixing address errors is a per process option.  The option is inherited
 * across fork(2) and execve(2) calls.  If you really want to use the
 * option in your user programs - I discourage the use of the software
 * emulation strongly - use the following code in your userland stuff:
 *
 * #include <sys/sysmips.h>
 *
 * ...
 * sysmips(MIPS_FIXADE, x);
 * ...
 *
 * The parameter x is 0 for disabeling software emulation.  Set bit 0 for
 * enabeling software emulation and bit 1 for enabeling printing debug
 * messages into syslog to aid finding address errors in programs.
 *
 * The logging feature is a addition over RISC/os and IRIX where only the
 * values 0 and 1 are acceptable values for x.  I'll probably remove this
 * hack later on.
 *
 * Below a little program to play around with this feature.
 *
 * #include <stdio.h>
 * #include <asm/sysmips.h>
 * 
 * struct foo {
 *         unsigned char bar[8];
 * };
 *
 * main(int argc, char *argv[])
 * {
 *         struct foo x = {0, 1, 2, 3, 4, 5, 6, 7};
 *         unsigned int *p = (unsigned int *) (x.bar + 3);
 *         int i;
 *
 *         if (argc > 1)
 *                 sysmips(MIPS_FIXADE, atoi(argv[1]));
 *
 *         printf("*p = %08lx\n", *p);
 *
 *         *p = 0xdeadface;
 *
 *         for(i = 0; i <= 7; i++)
 *         printf("%02x ", x.bar[i]);
 *         printf("\n");
 * }
 *
 * TODO:  complete the emulation of cp1/cp2 loads.
 */
#include <linux/mm.h>
#include <linux/signal.h>
#include <asm/inst.h>
#include <asm/segment.h>
#include <asm/unaligned.h>

inline int compute_return_epc(struct pt_regs *regs, char *exc);
/*
 * Damned ...  There is just too much bad code with that Intel desease in
 * it out there.   So we really need to support this :-(
 */
static __inline__ __uregister_t
load_mem(unsigned long addr, unsigned int size)
{
	switch(size) {
	case 2: return ldw_u((unsigned short *) addr);
	case 4: return ldl_u((unsigned int *) addr);
#if __mips >=3
	case 8: return ldq_u((unsigned long long *) addr);
#endif
	}

	/* Unreached */
	return 0;
}

static __inline__ void
store_mem(unsigned long addr, unsigned int size, __uregister_t val)
{
	switch(size) {
	case 2: stw_u(val, (unsigned short *) addr);
		return;
	case 4: stl_u(val, (unsigned int *) addr);
		return;
#if __mips >=3
	case 8: stq_u(val, (unsigned long long *) addr);
		return;
#endif
	}
}

/*
 * Read the content of some coprozessor register.
 * The implementation looks a bit stupid but this is the fastest and
 * simplest way.
 */
static __inline__ __uregister_t
read_cp_reg(int cpid, int cpreg)
{
	__uregister_t	result;

	result = 0;
	send_sig(SIGBUS, current, 1);
	return result;
}

static __inline__ void
write_cp_reg(int cpid, int cpreg, __uregister_t val)
{
	send_sig(SIGBUS, current, 1);
}

static __inline__ void
emulate_load_store_insn(struct pt_regs *regs, unsigned long addr, unsigned long pc)
{
	union mips_instruction insn;
	int errno, size, signop, cop;
	__register_t value;

	/*
	 * Assumes we're not in a branch delay slot.
	 */
	insn.word = get_user((unsigned int *)pc);
	switch (insn.i_format.opcode) {
	/*
	 * These are instruction that a compiler doesn't generate.  We
	 * can assume therefore that the code is MIPS-aware and
	 * really buggy.  Emulating these instructions would break the
	 * semantics anyway.
	 */
	case ll_op:
	case lld_op:
	case sc_op:
	case scd_op:

	/*
	 * For these instructions the only way to create an address
	 * error is an attempted access to kernel/supervisor address
	 * space.
	 */
	case ldl_op:
	case ldr_op:
	case lwl_op:
	case lwr_op:
	case sdl_op:
	case sdr_op:
	case swl_op:
	case swr_op:
	case lb_op:
	case lbu_op:
	case sb_op:
		send_sig(SIGBUS, current, 1);
		return;

	/*
	 * The remaining opcodes are the ones that are really of interrest.
	 */
	case lh_op:	size = 2; signop = 1; goto do_reg_load;
	case lhu_op:	size = 2; signop = 0; goto do_reg_load;
	case lw_op:	size = 4; signop = 1; goto do_reg_load;
	case lwu_op:	size = 4; signop = 0; goto do_reg_load;
	case ld_op:	size = 8; signop = 1;
do_reg_load:
		errno = verify_area(VERIFY_READ, (void *)addr, size);
		if (errno < 0) {
			send_sig(SIGSEGV, current, 1);
			return;
		}
		value = load_mem(addr, size);
		if (!signop) {
			/* This is an unsigned load */
			if (size == 2) value &= 0xffffUL;
			else if (size == 4) value &= 0xffffffffUL;
		}
		regs->regs[insn.i_format.rt] = value;
		regs->regs[insn.i_format.rt] = value;
		return;
	case sh_op:	size = 2; goto do_reg_store;
	case sw_op:	size = 4; goto do_reg_store;
	case sd_op:	size = 8;
do_reg_store:
		errno = verify_area(VERIFY_WRITE, (void *)addr, size);
		if (errno < 0) {
			send_sig(SIGSEGV, current, 1);
			return;
		}
		store_mem(addr, size, regs->regs[insn.i_format.rt]);
		return;

	case lwc1_op:	size = 4; cop = 1; goto do_cop_load;
	case lwc2_op:	size = 4; cop = 2; goto do_cop_load;
	case ldc1_op:	size = 8; cop = 1; goto do_cop_load;
	case ldc2_op:	size = 8; cop = 2;
do_cop_load:
		errno = verify_area(VERIFY_READ, (void *)addr, size);
		if (errno < 0) {
			send_sig(SIGSEGV, current, 1);
			return;
		}
		value = load_mem(addr, size);
		/*
		 * TODO: write value into coprozessor register.
		 */
		write_cp_reg(cop, insn.i_format.rt, value);
		return;

	case swc1_op:	size = 4; cop = 1; goto do_cop_store;
	case swc2_op:	size = 4; cop = 2; goto do_cop_store;
	case sdc1_op:	size = 8; cop = 1; goto do_cop_store;
	case sdc2_op:	size = 8; cop = 2;
do_cop_store:
		errno = verify_area(VERIFY_WRITE, (void *)addr, size);
		if (errno < 0) {
			send_sig(SIGSEGV, current, 1);
			return;
		}
		/*
		 * TODO: get value from coprozessor register.
		 */
		value = read_cp_reg(cop, insn.i_format.rt);
		store_mem(addr, size, value);
		return;
	default:
		/*
		 * Pheeee...  We encountered an yet unknown instruction ...
		 */
		send_sig(SIGILL, current, 1);
	}
}

static void
fix_ade(struct pt_regs *regs, unsigned long pc, char *exc)
{
	/*
	 * Did we catch a fault trying to load an instruction?
	 */
	if (regs->cp0_badvaddr == pc) {
		/*
		 * This fault occured during an attempt to load an instruction.
		 */
		send_sig(SIGBUS, current, 1);
	}

	/*
	 * Ok, this wasn't a failed instruction load.  The CPU was capable of
	 * reading the instruction and faulted after this.  So we don't need
	 * to verify_area the address of the instrucion.  We still don't
	 * know whether the address used was legal and therefore need to do
	 * verify_area().  The CPU already did the checking for legal
	 * instructions for us, so we don't need to do this.
	 */
	emulate_load_store_insn(regs, regs->cp0_badvaddr, pc);
}

void do_ade(struct pt_regs *regs)
{
	__register_t logpc, pc = regs->cp0_epc;
	char *adels;

	adels = (((regs->cp0_cause & CAUSEF_EXCCODE) >> CAUSEB_EXCCODE) == 4) ? "adel" : "ades";
	if (!user_mode(regs)) {
		show_regs(regs);
		panic("Caught %s exception in kernel mode", adels);
	}
	if (current->tss.mflags & MF_LOGADE) {
		logpc = pc;
		if (regs->cp0_cause & CAUSEF_BD)
			logpc += 4;
#if __mips >= 3
		printk(KERN_DEBUG
		       "Caught %s in '%s' at 0x%016Lx accessing 0x%016Lx.\n",
		       adels, current->comm, logpc, regs->cp0_badvaddr);
#else
		printk(KERN_DEBUG
		       "Caught %s in '%s' at 0x%08lx accessing 0x%08lx.\n",
		       adels, current->comm, logpc, regs->cp0_badvaddr);
#endif
	}
	if (compute_return_epc(regs, adels))
		return;
	if(current->tss.mflags & MF_FIXADE) {
		fix_ade(regs, pc, adels);
		return;
	}
#ifdef CONF_DEBUG_EXCEPTIONS
	show_regs(regs);
#endif
	send_sig(SIGBUS, current, 1);
}
